"use strict";

var _prototypeProperties = function (child, staticProps, instanceProps) { if (staticProps) Object.defineProperties(child, staticProps); if (instanceProps) Object.defineProperties(child.prototype, instanceProps); };

var _classCallCheck = function (instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } };

require("babel/polyfill");
var through = require("through2");
var detective = require("detective");
var combine = require("stream-combiner");
var path = require("path");
var _require = require("./f");

var binder = _require.binder;
var invoke = _require.invoke;
var not = _require.not;
var maskFilter = _require.maskFilter;
var catcher = _require.catcher;
var values = _require.values;
var getter = _require.getter;
var async = require("async");
var fs = require("fs");
var vinylFs = require("vinyl-fs");
var dependencyResolver = require("./dependencyResolver");
var VinylFile = require("vinyl");
var StreamConcat = require("stream-concat");
var beautify = require("js-beautify");
var bundleSerializer = require("./bundleSerializer");
var countDownLatch = require("./countDownLatch");
var UnresolvedDependenciesError = require("./errors/UnresolvedDependenciesError");
var crypto = require("crypto");
var VinylCache = require("./VinylCache");

//Very simple ad hoc profiling implementation. I couldn't find any good non-continuous profiling
//libraries in a couple of minutes, so writing my own is faster here. (suggestions for better frameworks
//are welcome!)
var startTimer = function () {
	return { stop: function stop() {} };
};
var printTimings = function () {};
if (process.env.YOLOADER_PROFILE) {
	(function () {
		var totalTimings = {};
		startTimer = function (name) {
			var start = new Date().getTime();
			return {
				stop: function stop() {
					var now = new Date().getTime();
					if (!start) {
						console.error("timer was already stopped");
					} else {
						var totalTiming = totalTimings[name] || { time: 0, calls: 0 };
						totalTiming.time += now - start;
						totalTiming.calls++;
						totalTimings[name] = totalTiming;
						start = null;
					}
				}
			};
		};
		printTimings = function () {
			console.log(totalTimings);
		};
	})();
}

var transformers = {

	compile: function compile(instance, compile) {
		return compile();
	},

	/**
  * Finds and attaches dependencies of a file
  */
	findDependencies: function findDependencies(instance) {
		instance.dependencyCache = instance.dependencyCache || new VinylCache();

		return through.obj(function (chunk, enc, done) {
			//If there are already deps set then the chunk probably comes from some cache. We won't bother
			//changing it then, it'll only mess up things.
			if (chunk.deps) {
				return done(null, chunk);
			}
			//We first try to get dependencies from cache.
			//Only if that fails we'll try to find them again
			var cachedDeps = instance.dependencyCache.get(chunk);

			if (cachedDeps) {
				chunk.deps = cachedDeps;
			} else {
				var timer = startTimer("findDeps");
				chunk.deps = {};

				detective(chunk.contents.toString())
				//For now all dependencies will have value null since they're not resolved yet
				.forEach(function (dep) {
					return chunk.deps[dep] = null;
				});

				//Update cache stuff
				instance.dependencyCache.set(chunk, chunk.deps);
				timer.stop();
			}
			done(null, chunk);
		});
	},

	/**
  * Resolves dependencies and adds them to the stream
  */
	resolveDependencies: function resolveDependencies(instance) {
		return through.obj(function resolveDependencies(chunk, enc, done) {
			var timer = startTimer("resolveDeps");
			//Push the current file, we'll need that anyway
			var onSuccess = catcher(done);

			var deps = Object.keys(chunk.deps);

			async.map(deps, dependencyResolver(chunk, instance), onSuccess(function (resolvedDeps) {
				//Check if we've found all dependencies:
				//At this point resolvedDeps should be an array of {file, base} pairs, but any dependency that could
				//not be resolved will have falsy values for the file values. We use that to filter out
				//unresolved dependencies.
				var unresolvedMask = resolvedDeps.map(getter("file")).map(not);
				var unresolved = deps.filter(maskFilter(unresolvedMask));
				if (unresolved.length) {
					return done(new UnresolvedDependenciesError(chunk.path, unresolved));
				}

				//Now we can attach the dependency objects to the chunk.deps object.
				resolvedDeps.forEach(function (dep, index) {
					var depName = deps[index];
					chunk.deps[depName] = dep;
				});

				timer.stop();
				done(null, chunk);
			}));
		});
	},

	compileDependencies: function compileDependencies(instance, compile, resolver) {
		//TODO: more efficient duplicate checking
		return through.obj(function (chunk, enc, done) {
			var timer = startTimer("compileDepsPrepare");

			var outer = this;
			this.push(chunk, enc);
			var newFiles = values(chunk.deps);
			var latch = countDownLatch(newFiles.length, function () {
				return done();
			});
			timer.stop();
			newFiles.forEach(function (file) {
				var timer = startTimer("compileDepsTrigger");
				vinylFs.src(file.file, { base: file.base }).pipe(resolver.resolveDependencies()).pipe(through.obj(function (chunk, enc, cb) {
					outer.push(chunk);
					//IMPORTANT: if we push the chunk here to the inner stream stuff blows up.
					//I don't know why (yet), but just don't do it
					cb();
				}, function (cb) {
					latch.countDown();cb();
				}));

				timer.stop();
			});
		});
	},

	/**
  * Resolves the require paths to links to actual files in the bundle.
  */
	linkDependencies: function linkDependencies(instance) {
		return through.obj(function (chunk, enc, done) {
			var timer = startTimer("link");
			Object.keys(chunk.deps).forEach(function (depName) {
				var dep = chunk.deps[depName];
				//If we have an aliased file, we'll use that name
				if (dep.as) {} else if (dep.file.startsWith(chunk.base)) {
					//If the dependency is in the same module as the file that requires it we can just
					//use a relative require. Otherwise we'll have to try a path-require or an external require
					chunk.deps[depName].as = "./" + path.relative(path.dirname(chunk.path), dep.file);
				} else {
					(function () {
						var pathEntry = instance.options.path.find(function (pathEntry) {
							return dep.file.startsWith(pathEntry.path);
						});
						if (!pathEntry) {
							throw new Error("Could not find the base module for " + dep.file);
						}
						chunk.deps[depName].as = pathEntry.name + "/" + path.relative(pathEntry.path, dep.file);
					})();
				}
			});
			timer.stop();
			done(null, chunk);
		});
	},

	/**
  * Bundles all streams together into one bundle object.
  */
	bundleStream: function bundleStream(instance, bundleOpts) {
		var bundle = {};
		/**
   * Gets (or creates) an object for a package.
   *
   * The object returned will be already created in the bundle object, so anything placed here
   * will end up in the bundle.
   */
		function getPackageObject(packageName) {
			if (!bundle[packageName]) {
				bundle[packageName] = {
					files: {},
					entry: []
				};
			}
			return bundle[packageName];
		}

		/**
   * Returns the name of a package based on it's path.
   *
   * For now it just implements path.basename, but in the future it might do something more intelligent,
   * like searching for the package.json.
   */
		function getPackageNameFromPath(filePath) {
			return path.basename(filePath);
		}

		//We'll bundle all the individual stream items into one object here
		return through.obj(function toBundle(chunk, enc, done) {
			var timer = startTimer("bundle");
			//First try to check if the vinyl base is exactly equal to some path entry. If it is, then we'll
			//use that as the package.
			var packageName = undefined;
			var pathEntry = instance.options.path.find(function (current) {
				return path.relative(chunk.base, current.path) === "";
			});
			if (pathEntry && pathEntry.name) {
				packageName = pathEntry.name;
			} else {
				packageName = getPackageNameFromPath(chunk.base);
			}
			var packageObject = getPackageObject(packageName);

			var filePath = undefined;
			var files = undefined;
			//If an explicit name is set we'll use that
			if (chunk.name) {
				var nameParts = chunk.name.split("/");
				filePath = nameParts.slice(1).join("/");
				packageName = nameParts[0];
				packageObject = getPackageObject(packageName);
				files = packageObject.files;
			} else if (chunk.path.startsWith(chunk.base)) {
				//If file is subpath of base then we can use a normal, relative require
				filePath = path.relative(chunk.base, chunk.path);
				files = packageObject.files;
			} else {
				throw new Error("Could not resolve file");
			}
			//We need to get hold of an object where we can place the content of the module.
			//Since the bundle is structured as an object representation of a file system we need to
			//walk through the bundle tree as if it was a file system.
			//With a combination of split and reduce we can walk through the bundle tree, creating path
			//entries on the fly, and finally end up at the object we're interested in.
			var target = filePath.split("/")
			//leading or trailing slashes will leave empty strings, so we use an identity function to
			//filter those out.
			.filter(function (x) {
				return x;
			}).reduce(function (bundle, part) {
				if (!bundle[part]) {
					bundle[part] = {};
				}
				return bundle[part];
			}, files);

			if (bundleOpts.entries.indexOf(chunk.path) !== -1) {
				packageObject.entry.push("./" + filePath);
			}

			//We've found our object, so we can place our info here
			//We keep the vinyl stream as it is, such that the serializer can later transform that into
			//an actual function
			target.content = chunk;
			target.deps = {};
			Object.keys(chunk.deps).forEach(function (dep) {
				target.deps[dep] = chunk.deps[dep].as;
			});
			done();
			timer.stop();
		}, function finalizeBundle(cb) {
			//NOTE: don't use arrow functions here, it binds this and messes stuff up
			this.push(bundle);
			cb();
		});
	},
	/**
  * Serializes a bundle object into a js file
  */
	serialize: function serialize(instance, bundleOpts) {
		return through.obj(function serialize(chunk, enc, done) {
			var timer = startTimer("serialize");
			done(null, bundleSerializer(chunk, instance, bundleOpts));
			timer.stop();
		});
	},
	/**
  *  Beautifies a javascript object if dev is set.
  *  It's just for debugging purposes and should not be used for actual module generation.
  *  It's not guaranteed to work with all features (e.g. atm it breaks sourcemaps)
  */
	beautify: (function (_beautify) {
		var _beautifyWrapper = function beautify() {
			return _beautify.apply(this, arguments);
		};

		_beautifyWrapper.toString = function () {
			return _beautify.toString();
		};

		return _beautifyWrapper;
	})(function (instance) {
		if (!instance.options.dev) {
			return through.obj();
		} else {
			return through.obj(function (chunk, enc, done) {
				var timer = startTimer("beautify");
				var beauty = beautify(chunk.contents.toString());
				chunk.contents = new Buffer(beauty);
				done(null, chunk);
				timer.stop();
			});
		}
	}),
	/**
  * Transforms a simple stream into a vinyl object.
  */
	createVinylStream: function createVinylStream(instance, bundleOpts) {
		return through.obj(function createVinylStream(chunk, enc, done) {
			done(null, new VinylFile({ contents: new Buffer(chunk), path: bundleOpts.name }));
		});
	}
};


var dependencyPipeline = [transformers.compile, transformers.findDependencies, transformers.resolveDependencies, transformers.compileDependencies];

var bundlePipeline = [transformers.linkDependencies, transformers.bundleStream, transformers.serialize, transformers.beautify];

function createPipeline(transformers) {
	for (var _len = arguments.length, opts = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
		opts[_key - 1] = arguments[_key];
	}

	return combine(transformers.map(binder.apply(undefined, opts)).map(invoke));
}

var Resolver = (function () {
	function Resolver(yoloaderInstance, compiler) {
		_classCallCheck(this, Resolver);

		this.yoloader = yoloaderInstance;
		this.compiler = compiler;
		this.filesProcessed = new Set();
	}

	_prototypeProperties(Resolver, null, {
		resolveDependencies: {
			value: function resolveDependencies() {
				var id = Math.random();
				var processor = this.yoloader.dependencyProcessor(this.yoloader, this.compiler, this);

				var self = this;
				//Filter the calls such that we only ever compile each file once
				var filter = through.obj(function (chunk, enc, done) {
					if (self.filesProcessed.has(chunk.path)) {
						return done(null, null);
					} else {
						self.filesProcessed.add(chunk.path);
						return done(null, chunk);
					}
				});
				return combine(filter, processor);
			},
			writable: true,
			configurable: true
		}
	});

	return Resolver;
})();

var Yoloader = (function () {
	function Yoloader() {
		var options = arguments[0] === undefined ? {} : arguments[0];
		_classCallCheck(this, Yoloader);

		this.dependencyProcessorPipeline = dependencyPipeline;
		this.bundlePipeline = bundlePipeline;
		this.dependencyProcessor = options.dependencyProcessor || createPipeline.bind(null, this.dependencyProcessorPipeline);
		this.bundler = options.bundler || createPipeline.bind(null, this.bundlePipeline);
		this.filesCompiled = [];
		this.filesPending = [];
		//We want an array of path objects, where each object is a { path, name } pair.
		//This is because at runtime we don't have full paths to resolve to, so we'll have to do it
		//name based. However, for convenience we also allow just path strings, from which we'll derive
		//the name by using path.basename on the full path. Here we convert all those strings to objects,
		//so that we don't need to bother about it later on.
		options.path = options.path || [];
		options.path = options.path.map(function (pathDef) {
			if (pathDef instanceof String || typeof pathDef === "string") {
				return { path: pathDef, name: path.basename(pathDef) };
			} else {
				return pathDef;
			}
		});
		this.path = options.path;

		this.mappings = options.mappings || {};
		this.dependencyResolvers = dependencyResolver.defaultResolvers;

		this.options = options;
	}

	_prototypeProperties(Yoloader, null, {
		resolveDependencies: {
			value: function resolveDependencies(compiler) {
				compiler = compiler || function (stream) {
					return stream;
				};
				return new Resolver(this, compiler).resolveDependencies();
			},
			writable: true,
			configurable: true
		},
		bundle: {
			value: function bundle(bundleOpts) {
				return this.bundler(this, bundleOpts);
			},
			writable: true,
			configurable: true
		}
	});

	return Yoloader;
})();

module.exports = Yoloader;

process.on("exit", printTimings);
//nothing to do